/**********************************************
 * comp.y - Yacc file for CS-218 Assignment 2 *
 * by Sean Handley, 320097@swan.ac.uk         *
 **********************************************/

/* declarations */
%{
#include <stdio.h>
#include "globals.h"
#include "y.tab.h"

#define YYSTYPE TreeNode *
static TreeNode *rootNode; /* keep a reference to the root node */

%}

%token ENDFILE ID BEG END VAR IF THEN ELSE FI
%token COMMA SEMI ASSIGN OPAREN CPAREN EQ NE LT GT LE GE
%token WHILE DO OD SKIP READ WRITE OR AND NOT
%token PLUS MINUS MULT DIV MOD TRUE FALSE NUMBER ERR

%%

/* grammar */

topprogram		: program { rootNode = $1; }

program 		: BEG declaration command_list END
					{
						$$ = newcmdnode(ProgT);
						$$->child[0] = $2;
						$$->child[1] = $3;
					}
				| BEG command_list END
					{
						$$ = $2;
					}
				;

declaration		: VAR id_list
					{
						$$ = newdecnode(DecT);
						$$->child[0] = $2;
					}
				;

id_list			: ID { $$ = $1; }
				| ID COMMA id_list
					{
						YYSTYPE t = $3;
						if (t != NULL) {
							while(t->sibling != NULL) {
								t = t->sibling;
							}
							t->sibling = $3;
						} else {
							$$ = $3;
						}
					}
				;

command_list	: command { $$ = $1; }
				| command_list SEMI command
					{
						YYSTYPE t = $1;
						if(t != NULL) {
							while(t->sibling != NULL) {
								t = t->sibling;
							}
							t->sibling = $3;
							$$ = $1;
						} else {
							$$ = $3;
						}
					}
				;

command			: ass_statement { $$ = $1; }
				| cond_statement { $$ = $1; }
				| iter_statement { $$ = $1; }
				| null_statement { $$ = $1; }
				| read_statement { $$ = $1; }
				| write_statement { $$ = $1; }
				;

ass_statement	: ID ASSIGN exp
					{
						$$ = newcmdnode(AssT);
						$$->child[0] = $3;
						$$->attr.name = "";
					}
				;

cond_statement	: IF comp THEN command_list ELSE command_list FI
					{
						$$ = newcmdnode(CondT);
						$$->child[0] = $2;
						$$->child[1] = $4;
						$$->child[2] = $6;
					}
				;

iter_statement	: WHILE comp DO command_list OD
					{
						$$ = newcmdnode(IterT);
						$$->child[0] = $2;
						$$->child[1] = $4;
					}
				;

null_statement	: SKIP { $$ = $1; }
				;

read_statement	: READ OPAREN ID CPAREN
					{
						$$ = newcmdnode(ReadT);
						$$->attr.name = "";
					}
				;

write_statement : WRITE OPAREN ID CPAREN
					{
						$$ = newcmdnode(WriteT);
						$$->attr.name = "";
					}
				;

comp			: bool_exp { $$ = $1; }
				| exp EQ exp
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $1;
						$$->child[1] = $3;
						$$->attr.op = EQ;
					}
				| exp NE exp
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $1;
						$$->child[1] = $3;
						$$->attr.op = NE;
					}
				| exp LT exp
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $1;
						$$->child[1] = $3;
						$$->attr.op = LT;
					}
				| exp GT exp
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $1;
						$$->child[1] = $3;
						$$->attr.op = GT;
					}
				| exp LE exp
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $1;
						$$->child[1] = $3;
						$$->attr.op = LE;
					}
				| exp GE exp
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $1;
						$$->child[1] = $3;
						$$->attr.op = GE;
					}
				;

bool_exp		: bool_term { $$ = $1; }
				| bool_exp OR bool_term
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $1;
						$$->child[1] = $3;
						$$->attr.op = OR;
					}
				;

bool_term		: bool_fact { $$ = $1; }
				| bool_term AND bool_fact
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $1;
						$$->child[1] = $3;
						$$->attr.op = AND;
					}
				;

bool_fact		: bool_atom { $$ = $1; }
				| OPAREN comp CPAREN { $$ = $2; }
				| NOT bool_fact
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $2;
						$$->attr.op = NOT;
					}
				;

exp				: term { $$ = $1; }
				| exp PLUS term
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $1;
						$$->child[1] = $3;
						$$->attr.op = PLUS;
					}
				| exp MINUS term
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $1;
						$$->child[1] = $3;
						$$->attr.op = MINUS;
					}
				;

term			: fact { $$ = $1; }
				| term MULT fact
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $1;
						$$->child[1] = $3;
						$$->attr.op = MULT;
					}
				| term DIV fact
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $1;
						$$->child[1] = $3;
						$$->attr.op = DIV;
					}
				| term MOD fact
					{
						$$ = newcmdnode(OpT);
						$$->child[0] = $1;
						$$->child[1] = $3;
						$$->attr.op = MOD;
					}
				;

fact			: atom { $$ = $1; }
				| OPAREN exp CPAREN { $$ = $2; }
				;

bool_atom		: TRUE
					{
						$$ = newcmdnode(OpT);
						$$->attr.op = TRUE;
					}
				| FALSE
					{
						$$ = newcmdnode(OpT);
						$$->attr.op = FALSE;
					}
				;

atom			: ID
					{
						$$ = newcmdnode(VarT);
						$$->attr.name = "";
					}
				| NUMBER
					{
						$$ = newcmdnode(VarT);
						$$->attr.val = 0;
					}
				;
				
%%

/* auxiliary functions */

int yyerror(const char *str) {
	fprintf(stderr,"error: %s\n",str);
}
 
int yywrap() {
	return 1;
} 

TreeNode *parse() {
	yyparse();
	return rootNode;
}
